@@ -1,6 +1,7 @@
 # noise.py
 #
-# This file is part of scqubits.
+# This file is part of scqubits: a Python package for superconducting qubits,
+# arXiv:2107.08552 (2021). https://arxiv.org/abs/2107.08552
 #
 #    Copyright (c) 2019 and later, Jens Koch and Peter Groszkowski
 #    All rights reserved.
@@ -178,10 +179,10 @@ class NoisySystem(ABC):
 
         # if we only have a single noise channel to consider (and hence are given a
         # str), put it into a one element list
-        noise_channels = (
-            [noise_channels] if isinstance(noise_channels, str) else noise_channels
+        noise_channels = cast(
+            List,
+            ([noise_channels] if isinstance(noise_channels, str) else noise_channels),
         )
-        cast(List, noise_channels)
 
         if spectrum_data is None:
             # We have to figure out the largest energy level involved in the
@@ -196,7 +197,7 @@ class NoisySystem(ABC):
                     opts = noise_channel[1]
                     max_level = max(max_level, opts.get("i", 1), opts.get("j", 1))
 
-            spectrum_data = self.get_spectrum_vs_paramvals(
+            spectrum_data = self.get_spectrum_vs_paramvals(  # type:ignore
                 param_name,  # type: ignore
                 param_vals,
                 evals_count=max_level + 1,
@@ -240,7 +241,7 @@ class NoisySystem(ABC):
         # remember current value of param_name
         current_val = getattr(self, param_name)
 
-        for n, noise_channel in enumerate(noise_channels):
+        for n, noise_channel in enumerate(noise_channels):  # type:ignore
 
             # case 1: noise_channel is a string representing the noise method
             if isinstance(noise_channel, str):
@@ -248,17 +249,21 @@ class NoisySystem(ABC):
                 noise_channel_method = noise_channel
 
                 # calculate the noise over the full param span in param_vals
-                noise_vals = [
-                    scale
-                    * getattr(self.set_and_return(param_name, v), noise_channel_method)(
-                        esys=(
-                            spectrum_data.energy_table[v_i, :],
-                            spectrum_data.state_table[v_i],
-                        ),
-                        **common_noise_options
-                    )
-                    for v_i, v in enumerate(param_vals)
-                ]
+                noise_vals = np.asarray(
+                    [
+                        scale
+                        * getattr(
+                            self.set_and_return(param_name, v), noise_channel_method
+                        )(
+                            esys=(
+                                spectrum_data.energy_table[v_i, :],  # type:ignore
+                                spectrum_data.state_table[v_i],  # type:ignore
+                            ),
+                            **common_noise_options
+                        )
+                        for v_i, v in enumerate(param_vals)
+                    ]
+                )
 
             # case 2: noise_channel is a tuple representing the noise method and
             # default options
@@ -273,17 +278,21 @@ class NoisySystem(ABC):
                 options.update(noise_channel[1])
 
                 # calculate the noise over the full param span in param_vals
-                noise_vals = [
-                    scale
-                    * getattr(self.set_and_return(param_name, v), noise_channel_method)(
-                        esys=(
-                            spectrum_data.energy_table[v_i, :],
-                            spectrum_data.state_table[v_i],
-                        ),
-                        **options
-                    )
-                    for v_i, v in enumerate(param_vals)
-                ]
+                noise_vals = np.asarray(
+                    [
+                        scale
+                        * getattr(
+                            self.set_and_return(param_name, v), noise_channel_method
+                        )(
+                            esys=(
+                                spectrum_data.energy_table[v_i, :],  # type:ignore
+                                spectrum_data.state_table[v_i],  # type:ignore
+                            ),
+                            **options
+                        )
+                        for v_i, v in enumerate(param_vals)
+                    ]
+                )
 
             else:
                 raise ValueError(
@@ -403,7 +412,7 @@ class NoisySystem(ABC):
                     opts = noise_channel[1]
                     max_level = max(max_level, opts.get("i", 1), opts.get("j", 1))
 
-            spectrum_data = self.get_spectrum_vs_paramvals(
+            spectrum_data = self.get_spectrum_vs_paramvals(  # type:ignore
                 param_name,
                 param_vals,
                 evals_count=max_level + 1,  # type: ignore
@@ -417,21 +426,23 @@ class NoisySystem(ABC):
         current_val = getattr(self, param_name)
 
         # calculate the noise over the full param span in param_vals
-        noise_vals = [
-            scale
-            * self.set_and_return(param_name, v).t1_effective(  # type: ignore
-                noise_channels=noise_channels,
-                common_noise_options=common_noise_options,
-                esys=(
-                    spectrum_data.energy_table[v_i, :],
-                    spectrum_data.state_table[v_i],
-                ),
-            )
-            for v_i, v in enumerate(param_vals)
-        ]
+        noise_vals = np.asarray(
+            [
+                scale
+                * self.set_and_return(param_name, v).t1_effective(  # type: ignore
+                    noise_channels=noise_channels,
+                    common_noise_options=common_noise_options,
+                    esys=(
+                        spectrum_data.energy_table[v_i, :],  # type:ignore
+                        spectrum_data.state_table[v_i],  # type:ignore
+                    ),
+                )
+                for v_i, v in enumerate(param_vals)
+            ]
+        )
 
         # Set the parameter we varied to its initial value
-        setattr(self, param_name, current_val)
+        setattr(self, param_name, current_val)  # type:ignore
 
         plotting_options = {
             "fig_ax": plt.subplots(1),
@@ -551,7 +562,7 @@ class NoisySystem(ABC):
                     opts = noise_channel[1]
                     max_level = max(max_level, opts.get("i", 1), opts.get("j", 1))
 
-            spectrum_data = self.get_spectrum_vs_paramvals(
+            spectrum_data = self.get_spectrum_vs_paramvals(  # type:ignore
                 param_name,
                 param_vals,
                 evals_count=max_level + 1,  # type: ignore
@@ -565,19 +576,21 @@ class NoisySystem(ABC):
         current_val = getattr(self, param_name)
 
         # calculate the noise over the full param span in param_vals
-        noise_vals = [
-            scale
-            * self.set_and_return(param_name, v).t2_effective(  # type: ignore
-                noise_channels=noise_channels,
-                common_noise_options=common_noise_options,
-                esys=(
-                    spectrum_data.energy_table[v_i, :],
-                    spectrum_data.state_table[v_i],
-                ),
-                get_rate=get_rate,
-            )
-            for v_i, v in enumerate(param_vals)
-        ]
+        noise_vals = np.asarray(
+            [
+                scale
+                * self.set_and_return(param_name, v).t2_effective(  # type: ignore
+                    noise_channels=noise_channels,
+                    common_noise_options=common_noise_options,
+                    esys=(
+                        spectrum_data.energy_table[v_i, :],  # type:ignore
+                        spectrum_data.state_table[v_i],  # type:ignore
+                    ),
+                    get_rate=get_rate,
+                )
+                for v_i, v in enumerate(param_vals)
+            ]
+        )
 
         # Set the parameter we varied to its initial value
         setattr(self, param_name, current_val)
